import re
from .base import MarkerBase, MarkerWithEnd, MarkerNoTranslator, cmd_call_prefix_simple
from .function import CallMarker
from .timout import TimeoutMarker
from .invoke import MarkerInvokerBase, InvokeMarker, GotoMarker


class MarkerAssignBase(MarkerBase):
    tag_head_re = r"[^\W\d]\w*[ \t\f\r\n]*%s"


class MarkerAssignValueBase(MarkerAssignBase):
    def execute(self, context, command, marker_node, marker_set):
        command = command.lstrip()
        index = re.match(self.tag_head_re % '', command).span()[1]
        var_name = command[:index].rstrip()
        rest = command[index:]
        index = re.match(self.tag_head_re_args, rest).span()[1]
        return self.assign_value(context, marker_node, marker_set, var_name, rest[index:].strip())

    def assign_value(self, context, marker_node, marker_set, var_name, expression):
        self.set_var_raw(context, var_name, self.get_value(context, marker_node, marker_set, expression))

    def get_value(self, context, marker_node, marker_set, expression):
        pass


class IgnoreMarker(MarkerBase):
    tag_head = ':'


class AssignStrMarker(MarkerAssignValueBase):
    tag_head_re_args = ':'

    def get_value(self, context, marker_node, marker_set, expression):
        return expression


class AssignRawStrMarker(AssignStrMarker, MarkerNoTranslator):
    tag_head_re_args = re.escape(r'\:')


class AssignMultiLineStrMarker(MarkerAssignBase, MarkerWithEnd):
    tag_head_re_args = '::'

    def execute(self, context, command, marker_node, marker_set):
        text = self.get_inner_content(context, marker_node)
        args = self.split_raw(command, 1, self.tag_head_re_args)
        self.set_var(context, args, 0, text)
        return []


class AssignMultiLineRawStrMarker(AssignMultiLineStrMarker, MarkerNoTranslator):
    tag_head_re_args = re.escape(r'\::')


class AssignEvalMarker(MarkerAssignValueBase, MarkerNoTranslator):
    tag_head_re_args = '='

    def get_value(self, context, marker_node, marker_set, expression):
        return self.eval(context, expression)


class AssignExecMarker(MarkerAssignBase, MarkerWithEnd, MarkerNoTranslator):
    tag_head_re_args = '=='

    def execute(self, context, command, marker_node, marker_set):
        code = self.get_inner_content(context, marker_node)
        result = self.eval_codes(context, code)
        args = self.split_raw(command, 1, self.tag_head_re_args)
        self.set_var(context, args, 0, result)
        return []


class AssignCmdcallMarker(MarkerAssignValueBase):
    tag_head_re_args = cmd_call_prefix_simple

    def get_value(self, context, marker_node, marker_set, expression):
        return self.eval_mixin(context, cmd_call_prefix_simple + expression, False)


class AssignCallMarker(MarkerAssignValueBase, CallMarker):
    tag_head = None
    tag_head_re_args = r'call[ \t\f\r\n]+'

    def assign_value(self, context, marker_node, marker_set, var_name, expression):
        def callback(value):
            self.set_var_raw(context, var_name, self.real_return_value(value))

        return *self.execute_core(context, expression), callback


class AssignMarkerInvokerBase(MarkerAssignValueBase, MarkerInvokerBase):
    tag_head = None

    def assign_value(self, context, marker_node, marker_set, var_name, expression):
        self.set_var_raw(context, var_name, self.execute_core(context, marker_set, expression))


class AssignInvokerMarker(AssignMarkerInvokerBase, InvokeMarker):
    tag_head_re_args = r'invoke[ \t\f\r\n]+'


class AssignGotoMarker(AssignMarkerInvokerBase, GotoMarker):
    tag_head_re_args = r'goto[ \t\f\r\n]+'


class AssignTimeoutMarker(MarkerAssignValueBase, TimeoutMarker):
    tag_head = None
    tag_head_re_args = r'timeout[ \t\f\r\n]+'

    def get_value(self, context, marker_node, marker_set, expression):
        timeout, expression = self.split_raw(expression, 1)
        return self.execute_core_timeout(context, expression, marker_node, marker_set, timeout)


class DelVarMarker(MarkerBase):
    tag_head = "del"

    def execute(self, context, command, marker_node, marker_set):
        name = self.split_raw(command, 1, self.tag_head)[-1]
        if name:
            context.remove_variable(name)
